/*
 * Copyright (c) 2008 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "velodyneConfig.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef HAVE_POCOLIBS
#include <portLib.h>
#endif
#include <velodyne/velodyneLib.h>
#include "velodyne_private.h"

/* Calibration data file, built from the XML data file provided 
   with each sensor */
#include "velodyneCalibration.h"

static void velodyne_init_calibration(void);

/*----------------------------------------------------------------------*/

/**
 * \brief Get access to the lidar device 
 *
 * name is either "udp" to read live packets from the sensor, or
 * the name of a file containing data captured using tcpdump
 */
VELODYNE_STR *
velodyne_open(char *name)
{
	VELODYNE_STR *dev;

	dev = (VELODYNE_STR *)malloc(sizeof(VELODYNE_STR));
	if (dev == NULL) {
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_NO_MEMORY);
#endif
		return NULL;
	}
	if (strcmp(name, "udp") == 0) {
		dev->fd = velodyne_setup_socket();
		if (dev->fd == -1) {
			free(dev);
			return NULL;
		}
		dev->type = VELODYNE_NETWORK;
	} else {
		if (velodyne_pcap_file_open(name, &dev->pcap) == -1) {
			free(dev);
			return NULL;
		}
		dev->type = VELODYNE_FILE;
	}
	velodyne_init_calibration();
	return dev;
}

/*----------------------------------------------------------------------*/

/**
 * \brief Terminate the access to the device
 */
void
velodyne_close(VELODYNE_STR *device)
{
	switch (device->type) {
	case VELODYNE_FILE:
		velodyne_close_pcap(device->pcap);
	case VELODYNE_NETWORK:
		velodyne_close_socket(device->fd);
#ifdef HAVE_POCOLIBS
	default:
		errnoSet(S_velodyneLib_INVALID_DEVICE);
#endif
	}
	free(device);
}

/*----------------------------------------------------------------------*/
/**
 * \brief Read a packet from the device 
 */
int
velodyne_read_packet(VELODYNE_STR *device, VELODYNE_PACKET_STR *p)
{
	switch (device->type) {
	case VELODYNE_FILE:
		return velodyne_pcap_read_packet(device->pcap, p);
	case VELODYNE_NETWORK:
		return velodyne_udp_rcv(device->fd, p, 0);
	default:
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_INVALID_DEVICE);
#endif
		return -1;
	}
	return 0;
}

/*----------------------------------------------------------------------*/

void
velodyne_print_packet(VELODYNE_PACKET_STR *p)
{
	int i, j;
	VELODYNE_RAW_PACKET_STR *packet = &p->data;
	struct tm localtime;
	char buf[100];

	localtime_r(&p->ts.tv_sec, &localtime);
	strftime(buf, sizeof(buf), "%F %T", &localtime);
	printf("%s.%06u\n", buf, (unsigned long)p->ts.tv_usec);
	for (i = 0; i < VELODYNE_NUM_SHOTS; i++) {
		printf("Block %d %s, %d:\n", i, 
		    packet->shots[i].lower_upper == VELODYNE_UPPER_HEADER ?
		    "upper" : 
		    packet->shots[i].lower_upper == VELODYNE_LOWER_HEADER ? 
		    "lower" : "unknown",
		    packet->shots[i].rotational_pos);
		printf("Distance: ");
		for (j = 0; j < VELODYNE_NUM_LASERS; j++)
			printf("%d\t", packet->shots[i].lasers[j].distance);
		printf("\nIntensity: ");
		for (j = 0; j < VELODYNE_NUM_LASERS; j++)
			printf("%d\t", packet->shots[i].lasers[j].intensity);
		printf("\n");
	}
	printf("Status: %f %s", packet->degrees + packet->frac_degrees/256.0,
	    packet->status_string);
	printf("\n\n");
}

/*----------------------------------------------------------------------*/

void
velodyne_print_packet_3d(VELODYNE_PACKET_STR *pkt)
{
	VELODYNE_RAW_PACKET_STR *packet = &pkt->data;
	struct VELODYNE_POINT p;
	struct VELODYNE_FIRE_STR *s;
	int i, j, n;

	for (i = 0; i < VELODYNE_NUM_SHOTS; i++) {
		s = &(packet->shots[i]);
		for (j = 0; j < VELODYNE_NUM_LASERS; j++) {
			n = (s->lower_upper == VELODYNE_UPPER_HEADER ? 1 : 0)
			    * VELODYNE_NUM_LASERS + j;
			velodyne_compute_coords(n, s->rotational_pos, 
			    s->lasers[j].distance, &p);
			printf("%lf %lf %lf %d\n", 
			    p.x, p.y, p.z, s->lasers[j].intensity);
		}
	}
}

/*----------------------------------------------------------------------*/

#define VLS_NUM_ROT_ANGLES  36000

double rotCosTable[VLS_NUM_ROT_ANGLES];
double rotSinTable[VLS_NUM_ROT_ANGLES];

static void
velodyne_init_calibration(void)
{
	int i;

	for (i = 0; i < VLS_NUM_ROT_ANGLES; i++) {
		rotCosTable[i] = cos(2*i*M_PI/VLS_NUM_ROT_ANGLES);
		rotSinTable[i] = sin(2*i*M_PI/VLS_NUM_ROT_ANGLES);
	}
}

/**
 * Function to convert time of flight information into 3D position.
 * This isn't particularly optimized, for clarity sake, other than
 * that most sine and cosine values have been precomputed.  The "db"
 * object holds the calibration data for each laser and the "pos"
 * object holds the current 3D position of the scanner (i.e. from
 * GPS).  This code doesn't factor in orientation of the scanner
 * (roll, pitch, yaw), but it is fairly straightforward to factor
 * these values in.  The physical offsets between the two laser blocks
 * is not currently taken into account, but adding these corrections
 * is fairly straightforward.  Please note that this code is
 * structured try and make it clear about how the various corrections
 * are applied and is not the most efficient implementation.
 * 
 * The following correction values are used:
 *
 * distLSB - the CM value for the lsb of a time of flight distance readbing
 * vertCorrection - the elevation angle for each laser 
 *                  (positive rotates the beam towards the top of the scanner) 
 * rotCorrection - the azimuth angle offset for each laser 
 *                 (offset from the current rotation angle, 
 *                 positive rotates the beam counter-clockwise)
 * distCorrection - a CM distance offset applied to the time 
 *                  of flight distance given by laser
 * horizOffsetCorrection - a horizontal parallax correction 
 *                         (orthogonal to laser beam)
 * vertOffsetCorrection - a vertical parallax correction 
 *                        (orthogonal to laser beam)
 *
 * *Note: If the beam were initially pointed directly along +y axis of
 * the world frame, a positive horizontal offset correction would
 * shift the beam to the left (-x) and a positive vertial offset
 * correction correction whould shift the beam up (+z)
 *
 * In these computations the positive Y-axis is at rotational degree
 * zero and the scanner rotates clockwise around the Z-axis.
 */


const double VLS_DIM_SCALE = 1.0;	/* ?? */

void
velodyne_compute_coords(u_int16_t laserNum, int position, int dist, 
    struct VELODYNE_POINT *point)
{
	u_int16_t idx = laserNum % VELODYNE_NUM_LASERS;
	struct VELODYNE_CALIBRATION cal = vc[idx];

	if (dist == 0) {
		point->x = 0.0;
		point->y = 0.0;
		point->z = 0.0;
		return;
	}

	double distance = velodyne_distLSB * dist + cal.distCorrection;

	double cosVertAngle = cos(cal.vertCorrection * M_PI / 180.0);
	double sinVertAngle = sin(cal.vertCorrection * M_PI / 180.0);
	double cosRotCorrection = cos(cal.rotCorrection * M_PI / 180.0);
	double sinRotCorrection = sin(cal.rotCorrection * M_PI / 180.0);

	/* cos(a-b) = cos(a)*cos(b) + sin(a)*sin(b) */
	/* sin(a-b) = sin(a)*cos(b) - cos(a)*sin(b) */
	double cosRotAngle = rotCosTable[position]*cosRotCorrection + rotSinTable[position]*sinRotCorrection;
	double sinRotAngle = rotSinTable[position]*cosRotCorrection - rotCosTable[position]*sinRotCorrection;

	distance /= VLS_DIM_SCALE;

	/* The offset corrections are to be applied in planes orthogonal to the rotatation corrected beam */
	double hOffsetCorr = cal.horizOffsetCorrection/VLS_DIM_SCALE;
	double vOffsetCorr = cal.vertOffsetCorrection/VLS_DIM_SCALE;

/*
 *                           / (distance, shifted by vertical offset)
 *    z ^                  /
 *      |                /
 *      --> y          /          distance
 *     /             /           /            vertOffsetCorrection
 *    v x          /           /            +-----------+
 *               /           /              |     ^     |
 *             /           /                |     |     |
 * vertOffset/           /                  |<----o     | horizOffsetCorrection
 *          \          /                    |           |
 *           \ 90deg /                      |           |
 *            \    /         xyDist         +-----------+
 *    90-theta \ /   theta   |              Note: the "o" represents the beam pointing into the screen
 *         ------------------------               if the beam were aligned with +y in the world frame
 *           ^  |         ^                       then vertOffset would be aligned with +z and hoizOffset
 *           |     distance*cos(theta)            would be aligned with -x
 *  vertOffset*cos(90-theta) = vertOffset*sin(theta)
 *
 * Note: theta = vertCorrection angle
 *
 *                   / (x,y)
 *                 /
 *     y         /
 *    ^        /           xyDist
 *    |      /   |       /
 *    -->x   \   |theta/
 *            \  |   /
 *   horizOff  \ | /
 *           ----/-------- xyDist*sin(theta)
 *
 * Note: theta = rotCorrection angle
 */
	// Compute the distance in the xy plane (without accounting for rotation)
	double xyDistance = distance * cosVertAngle - vOffsetCorr * sinVertAngle;

	// pos is the position of the scanner, factor in rotation angle and horizontal offset
	point->x = xyDistance * sinRotAngle - hOffsetCorr * cosRotAngle;
	point->y = xyDistance * cosRotAngle + hOffsetCorr * sinRotAngle;
	point->z = distance * sinVertAngle + vOffsetCorr * cosVertAngle;
}
